#include "string.h"

/* ESP drivers */
#include "esp_event.h"
#include "esp_log.h"
#include "esp_system.h"
#include "esp_wifi.h"

/* FreeRTOS */
#include "freertos/FreeRTOS.h"
#include "freertos/event_groups.h"
#include "freertos/task.h"

/* App specific */
#include "app_wifi.h"

#define LOG_TAG "APP_WIFI"

#define APP_WIFI_AUTH_MODE WIFI_AUTH_WPA2_PSK

#define APP_WIFI_EVENT_GROUP_EVENT_CONNECTED (1 << 0U)

EventGroupHandle_t s_app_wifi_event_group;

static uint8_t      s_retries = 0U;
static esp_netif_t *s_sta_netif;

static void app_wifi_event_handler(void *arg, esp_event_base_t event_base, int32_t event_id, void *event_data);

esp_err_t app_wifi_init(void) {
    s_app_wifi_event_group = xEventGroupCreate();
    if (s_app_wifi_event_group == NULL) {
        return ESP_FAIL;
    }

    ESP_ERROR_CHECK(esp_netif_init());
    ESP_ERROR_CHECK(esp_event_loop_create_default());
    s_sta_netif = esp_netif_create_default_wifi_sta();

    wifi_init_config_t wifi_init_cfg = WIFI_INIT_CONFIG_DEFAULT();
    ESP_ERROR_CHECK(esp_wifi_init(&wifi_init_cfg));

    esp_event_handler_instance_t inst_any_id;
    esp_event_handler_instance_t inst_got_ip;

    ESP_ERROR_CHECK(
        esp_event_handler_instance_register(WIFI_EVENT, ESP_EVENT_ANY_ID, app_wifi_event_handler, NULL, &inst_any_id));

    ESP_ERROR_CHECK(
        esp_event_handler_instance_register(IP_EVENT, IP_EVENT_STA_GOT_IP, app_wifi_event_handler, NULL, &inst_got_ip));

    wifi_config_t wifi_config;
    bool          has_changes = false;

    ESP_ERROR_CHECK(esp_wifi_get_config(WIFI_IF_STA, &wifi_config));

    if (strncmp((char *)wifi_config.sta.ssid, CONFIG_APP_WIFI_SSID, strlen(CONFIG_APP_WIFI_SSID)) != 0) {
        snprintf((char *)wifi_config.sta.ssid, 32, CONFIG_APP_WIFI_SSID);
        has_changes = true;
    }

    if (strncmp((char *)wifi_config.sta.password, CONFIG_APP_WIFI_PASSWORD, strlen(CONFIG_APP_WIFI_PASSWORD)) != 0) {
        snprintf((char *)wifi_config.sta.password, 64, CONFIG_APP_WIFI_PASSWORD);
        has_changes = true;
    }

    if (wifi_config.sta.threshold.authmode != APP_WIFI_AUTH_MODE) {
        wifi_config.sta.threshold.authmode = APP_WIFI_AUTH_MODE;
        has_changes                        = true;
    }

    ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_STA));

    if (has_changes) {
        ESP_ERROR_CHECK(esp_wifi_set_config(WIFI_IF_STA, &wifi_config));
        ESP_LOGI(LOG_TAG, "WiFi config has changed, setting new config.");
    }

    ESP_ERROR_CHECK(esp_wifi_start());

    ESP_LOGI(LOG_TAG, "WiFi initialized, heap free: %lu.", esp_get_free_heap_size());

    return ESP_OK;
}

esp_err_t app_wifi_deinit(void) {
    esp_wifi_stop();
    esp_wifi_deinit();

    esp_netif_destroy_default_wifi(s_sta_netif);

    esp_event_loop_delete_default();
    vEventGroupDelete(s_app_wifi_event_group);

    return ESP_OK;
}

app_wifi_event_t app_wifi_wait_event(uint32_t max_timeout_ms) {
    EventBits_t bits = xEventGroupWaitBits(s_app_wifi_event_group, APP_WIFI_EVENT_GROUP_EVENT_CONNECTED, pdFALSE,
                                           pdFALSE, pdMS_TO_TICKS(max_timeout_ms));

    if (bits & APP_WIFI_EVENT_GROUP_EVENT_CONNECTED) {
        return APP_WIFI_EVENT_CONNECTED;
    } else {
        return APP_WIFI_EVENT_TIMEOUT;
    }
}

static void app_wifi_event_handler(void *arg, esp_event_base_t event_base, int32_t event_id, void *event_data) {
    /* Event handler */

    if (event_base == WIFI_EVENT) {
        if (event_id == WIFI_EVENT_STA_START) {
            esp_wifi_connect();
        }

        if (event_id == WIFI_EVENT_STA_DISCONNECTED) {
            esp_wifi_connect();
            ESP_LOGI(LOG_TAG, "Disconnected");
        }
    } else if (event_base == IP_EVENT) {
        if (event_id == IP_EVENT_STA_GOT_IP) {
            xEventGroupSetBits(s_app_wifi_event_group, APP_WIFI_EVENT_GROUP_EVENT_CONNECTED);
            s_retries = 0U;

            ip_event_got_ip_t *event = event_data;

            ESP_LOGI(LOG_TAG, "Connected, IP address: " IPSTR, IP2STR(&event->ip_info.ip));
        }
    }
}
